--- old/usr/src/man/man2/unlink.2	Sun Dec 18 02:23:47 2016
+++ new/usr/src/man/man2/unlink.2	Sun Dec 18 02:23:47 2016
@@ -2,6 +2,7 @@
 .\" Copyright (c) 2007, Sun Microsystems, Inc.  All Rights Reserved.
 .\" Copyright 1989 AT&T
 .\" Portions Copyright (c) 1992, X/Open Company Limited  All Rights Reserved.
+.\" Copyright 2016 RackTop Systems.
 .\" Sun Microsystems, Inc. gratefully acknowledges The Open Group for permission to reproduce portions of its copyrighted documentation. Original documentation from The Open Group can be obtained online at
 .\" http://www.opengroup.org/bookstore/.
 .\" The Institute of Electrical and Electronics Engineers and The Open Group, have given us permission to reprint portions of their documentation. In the following statement, the phrase "this text" refers to portions of the system documentation. Portions of this text are reprinted and reproduced in electronic form in the Sun OS Reference Manual, from IEEE Std 1003.1, 2004 Edition, Standard for Information Technology -- Portable Operating System Interface (POSIX), The Open Group Base Specifications Issue 6, Copyright (C) 2001-2004 by the Institute of Electrical and Electronics Engineers, Inc and The Open Group. In the event of any discrepancy between these versions and the original IEEE and The Open Group Standard, the original IEEE and The Open Group Standard is the referee document. The original Standard can be obtained online at http://www.opengroup.org/unix/online.html.
@@ -9,7 +10,7 @@
 .\" The contents of this file are subject to the terms of the Common Development and Distribution License (the "License").  You may not use this file except in compliance with the License.
 .\" You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE or http://www.opensolaris.org/os/licensing.  See the License for the specific language governing permissions and limitations under the License.
 .\" When distributing Covered Code, include this CDDL HEADER in each file and include the License file at usr/src/OPENSOLARIS.LICENSE.  If applicable, add the following below this CDDL HEADER, with the fields enclosed by brackets "[]" replaced with your own identifying information: Portions Copyright [yyyy] [name of copyright owner]
-.TH UNLINK 2 "May 18, 2007"
+.TH UNLINK 2 "Dec 15, 2016"
 .SH NAME
 unlink, unlinkat \- remove directory entry
 .SH SYNOPSIS
@@ -61,12 +62,8 @@
 .sp
 .LP
 If the \fIpath\fR argument is a directory and the filesystem supports
-\fBunlink()\fR and \fBunlinkat()\fR on directories, the directory is unlinked
-from its parent with no cleanup being performed.  In UFS, the disconnected
-directory will be found the next time the filesystem is checked with
-\fBfsck\fR(1M). The \fBunlink()\fR and \fBunlinkat()\fR functions will not fail
-simply because a directory is not empty. The user with appropriate privileges
-can orphan a non-empty directory without generating an error message.
+\fBunlink()\fR and \fBunlinkat()\fR on directories the behaviour is dependent
+on the filesystem.
 .sp
 .LP
 If the \fIpath\fR argument is a directory and the filesystem does not support
--- old/usr/src/uts/common/fs/tmpfs/tmp_dir.c	Sun Dec 18 02:23:47 2016
+++ new/usr/src/uts/common/fs/tmpfs/tmp_dir.c	Sun Dec 18 02:23:47 2016
@@ -21,10 +21,10 @@
 /*
  * Copyright 2007 Sun Microsystems, Inc.  All rights reserved.
  * Use is subject to license terms.
+ *
+ * Copyright 2016 RackTop Systems.
  */
 
-#pragma ident	"%Z%%M%	%I%	%E% SMI"
-
 #include <sys/types.h>
 #include <sys/param.h>
 #include <sys/sysmacros.h>
@@ -525,11 +525,19 @@
 	dir->tn_ctime = now;
 	tp->tn_ctime = now;
 
+	/*
+	 * If this is a _REMOVE (unlink) operation there may
+	 * be other links to the directory entry.
+	 */
 	ASSERT(tp->tn_nlink > 0);
 	DECR_COUNT(&tp->tn_nlink, &tp->tn_tlock);
-	if (op == DR_RMDIR && tp->tn_type == VDIR) {
-		tdirtrunc(tp);
-		ASSERT(tp->tn_nlink == 0);
+	if (op == DR_RMDIR || (op == DR_REMOVE && tp->tn_type == VDIR)) {
+		if (tp->tn_nlink > 1) {
+			ASSERT(op == DR_REMOVE);
+		} else {
+			tdirtrunc(tp);
+			ASSERT(tp->tn_nlink == 0);
+		}
 	}
 	return (0);
 }
@@ -615,6 +623,7 @@
 	size_t namelen;
 	timestruc_t now;
 	int isvattrdir, isdotdot, skip_decr;
+	int lock_held;
 
 	ASSERT(RW_WRITE_HELD(&dir->tn_rwlock));
 	ASSERT(dir->tn_type == VDIR);
@@ -653,6 +662,29 @@
 		tmp_memfree(tdp, sizeof (struct tdirent) + namelen);
 		dir->tn_size -= (sizeof (struct tdirent) + namelen);
 		dir->tn_dirents--;
+
+		/*
+		 * This directory entry may itself be a directory with
+		 * entries and removing it may have created orphans.
+		 * On a normal filesystem like UFS this wouldn't be
+		 * a huge problem because fcsk can reclaim them.  For
+		 * TMPFS which resides in RAM however, it means we
+		 * end up leaking memory.
+		 *
+		 * To avoid this we also truncate child directories,
+		 * but only if they have no other links to them.
+		 */
+		if (!isdotdot && tp->tn_type == VDIR && tp != dir) {
+			if (tp->tn_nlink > 1)
+				continue;
+			lock_held = RW_WRITE_HELD(&tp->tn_rwlock);
+			if (!lock_held)
+				rw_enter(&tp->tn_rwlock, RW_WRITER);
+			tdirtrunc(tp);
+			if (!lock_held)
+				rw_exit(&tp->tn_rwlock);
+			ASSERT(tp->tn_nlink == 0);
+		}
 	}
 
 	gethrestime(&now);
--- old/usr/src/uts/common/fs/tmpfs/tmp_vnops.c	Sun Dec 18 02:23:47 2016
+++ new/usr/src/uts/common/fs/tmpfs/tmp_vnops.c	Sun Dec 18 02:23:47 2016
@@ -1105,8 +1105,7 @@
 
 	if (tp->tn_type != VDIR ||
 	    (error = secpolicy_fs_linkdir(cred, dvp->v_vfsp)) == 0)
-		error = tdirdelete(parent, tp, nm, tp->tn_type == VDIR ?
-		    DR_RMDIR : DR_REMOVE, cred);
+		error = tdirdelete(parent, tp, nm, DR_REMOVE, cred);
 
 	rw_exit(&tp->tn_rwlock);
 	rw_exit(&parent->tn_rwlock);
